import { IApp } from "../interfaces/IApp";
import { ICoherenceContainer } from "../../coherence/interfaces/ICoherenceContainer";
import { ContainerNotInitializedException } from "../../exceptions/coherence/ContainerNotInitializedException";
import { InMemoryContainer } from "../../coherence/concrete/InMemoryContainer";
import { IGeranium } from "../interfaces/IGeranium";
import { AjaxCommunicator } from "../../backend/concrete/ajax/AjaxCommunicator";
import { InstantiatedException } from "../../exceptions/runtime/InstantiatedException";
import { Html5HistoryAPI } from "../../history/concrete/Html5HistoryAPI";
import { ConsoleLogger } from "../../exceptions/logging/concrete/ConsoleLogger";
import { AjaxRequest } from "../../backend/concrete/ajax/AjaxRequest";
import { BasicRouter } from "../../routing/concrete/BasicRouter";
import { IStateManager } from "../../states/interfaces/IStateManager";
import { LocalStorage } from "../../storage/concrete/LocalStorage";
import { ClientTemplateEngine } from "../../templating/concrete/mustache/ClientTemplateEngine";
import { NotifyValidatingReporter } from "../../validating/reporter/concrete/NotifyValidatingReporter";
import { BaseViewBinder } from "../../viewbinding/concrete/BaseViewBinder";
import { BaseFieldBinding } from "../../binding/concrete/BaseFieldBinding";
import { BaseClickBinding } from "../../binding/concrete/BaseClickBinding";
import { BaseInputBinding } from "../../binding/concrete/BaseInputBinding";
import { BaseCollectionBinding } from "../../binding/concrete/BaseCollectionBinding";
import { ICommunicator } from "../../backend/interfaces/ICommunicator";
import { IHistory } from "../../history/interfaces/IHistory";
import { ILogger } from "../../exceptions/logging/interfaces/ILogger";
import { IRequest } from "../../backend/interfaces/IRequest";
import { IRouter } from "../../routing/interfaces/IRouter";
import { IStorage } from "../../storage/interfaces/IStorage";
import { ITemplateEngine } from "../../templating/interfaces/ITemplateEngine";
import { IValidatingReporter } from "../../validating/reporter/interfaces/IValidatatingReporter";
import { IViewBinder } from "../../viewbinding/interfaces/IViewBinder";
import { IBinding } from "../../binding/interfaces/IBinding";
import { IInjected } from "../../coherence/interfaces/IInjected";
import { IViewEngine } from "../../viewengine/interfaces/IViewEngine";
import { BaseViewEngine } from "../../viewengine/concrete/BaseViewEngine";
import { BaseViewPublisher } from "../../viewengine/concrete/BaseViewPublisher";
import { IViewPublisher } from "../../viewengine/interfaces/IViewPublisher";
import { PropertyBinding } from "../../binding/concrete/PropertyBinding";

class App implements IApp {
    ["`container"]: ICoherenceContainer;
    public static containerProperty: string = "`GeraniumApp";
    private instantiated: boolean = false;


    register<T extends IInjected>(type: Function | (new (...args: any[]) => T), value: T): void {
        return this["`container"].register(type, value);
    }
    resolve<T extends IInjected>(type: Function | (new (...args: any[]) => T)): T {
        return this["`container"].resolve(type);
    }
    resolveAll<T extends IInjected>(type: Function | (new (...args: any[]) => T)): T[] {
        return this["`container"].resolveAll(type);
    }
    release<T extends IInjected>(type: Function | (new (...args: any[]) => T)) {
        return this["`container"].release(type);
    }
    instantiate<T extends IInjected>(type: any, params?: any[]): T {
        return this["`container"].instantiate(type, params);
    }
    all(): any[] {
        return this["`container"].all();
    }

    start(geranium?: IGeranium) {
        if (this.instantiated) {
            throw new InstantiatedException("GeraniumApp.instantiate");
        }

        /** apply settings */
        Object.assign(geranium, geraniumDefault);

        /** apply container */
        this["`container"] = geranium.container ? new geranium.container() : new InMemoryContainer();

        /** register all settings */
        this.internalRegister(geranium);

        this.instantiated = true;
    }

    private internalRegister(geranium: IGeranium) {
        this["`container"].register(ICommunicator, new geranium.communicator());
        this["`container"].register(IHistory, new geranium.historyprovider());
        this["`container"].register(ILogger, new geranium.logger());
        this["`container"].register(IRequest, new geranium.request());
        this["`container"].register(IRouter, new geranium.router());
        this["`container"].register(IStateManager, new geranium.statemanager());
        this["`container"].register(IStorage, new geranium.storage());
        this["`container"].register(ITemplateEngine, new geranium.templating());
        this["`container"].register(IValidatingReporter, new geranium.validationreporter());
        this["`container"].register(IViewBinder, new geranium.viewbinder());
        this["`container"].register(IViewEngine, new geranium.viewengine());
        this["`container"].register(IViewPublisher, new geranium.viewpublisher());
        geranium.bindings.forEach(binding => {
            this["`container"].register(IBinding, new binding());
        });
    }
}

const geraniumDefault: IGeranium = {
    communicator: AjaxCommunicator,
    historyprovider: Html5HistoryAPI,
    logger: ConsoleLogger,
    request: AjaxRequest,
    router: BasicRouter,
    statemanager: InMemoryContainer,
    storage: LocalStorage,
    templating: ClientTemplateEngine,
    validationreporter: NotifyValidatingReporter,
    viewbinder: BaseViewBinder,
    viewengine: BaseViewEngine,
    viewpublisher: BaseViewPublisher,
    bindings: [
        BaseFieldBinding as any,
        BaseInputBinding,
        BaseClickBinding,
        BaseCollectionBinding,
        PropertyBinding
    ]
}

function getApp(): IApp {
    if (!window[App.containerProperty]) {
        window[App.containerProperty] = new App();
    }
    return window[App.containerProperty];
}

/** Application intance with in-build container */
var GeraniumApp = getApp();

/** Application intance with in-build container */
export default GeraniumApp;